#include "PlayState.h"
#include "cmPlatform.h"
#include "GGJ2015.h"
#include "InGameScreen.h"
#include <algorithm>
#include <cstdio>
#include "cmMath.h"


#include "GameStateManager.h"
#include "MainMenuState.h"

// Entities
#include "BlastDoor.h"
#include "FlameHazard.h"
#include "SpawnChute.h"

const float PlayState::ms_fMinimumScreenDragDistance = 30.0f;
const float PlayState::ms_fMaximumScreenDragDistance = 100.0f;
const float PlayState::ms_fVacuumAcceleration = 0.05f; // accelerating by units per second
const float PlayState::ms_fMaximumVacuumVelocity = 0.5f; // max units per second
const float PlayState::ms_fVacuumTimeMS = 15000.0f; // 15 seconds
const float PlayState::ms_fWinningTimeMs = 20000.0f; // 20 seconds
const int PlayState::ms_nMaxPunches = 20;
const int PlayState::ms_nMaxSpacemen = 3;

PlayState::PlayState()
{
}

PlayState::~PlayState()
{
}

void PlayState::OnStart()
{
	GGJ2015::Get()->GetCamera().Init();

	gAudio->PlayMusicLooping(L"Action BG.ogg", 0.4f);
	
	// Lets start with some spacemen
	RebuildLevel(true);
	EndTurn();
}

void PlayState::Update()
{
	GGJ2015::Get()->GetCamera().Update();
	cmScreenManager::Get().Update();

	m_fStageTimer += gPlatform->GetDeltaTime();
	if (m_currentPlayState == Respawn)
	{
		std::vector<Spaceman*> spacemen;
		GGJ2015::Get()->GetEntityManager()->GetEntites<Spaceman>(spacemen);
		
		int nNumSpaceman = 0;
		for (auto spaceman : spacemen)
		{
			if (spaceman->m_player == GGJ2015::Get()->m_playerTurn)
			{
				++nNumSpaceman;
			}
		}

		if (nNumSpaceman < ms_nMaxSpacemen)
		{
			RespawnSpaceman(GGJ2015::Get()->m_playerTurn);
		}

		SetState(Select);
	}

	if (m_currentPlayState == Launched)
	{
		if (GGJ2015::Get()->GetEntityManager()->IsSystemAtRest())
		{
			EndTurn();
		}
	}
	else if (m_currentPlayState == TransitionStage_Vaccuum)
	{
		SuckSpacemenTowardStageDoor();
		std::vector<Spaceman*> spacemen;
		GGJ2015::Get()->GetEntityManager()->GetEntites<Spaceman>(spacemen);

		// if we hit a timer or we run out of spacemen
		if (m_fStageTimer > ms_fVacuumTimeMS || spacemen.empty())
		{
			// Move the level
			if (m_nSuctionSample != 0)
				gAudio->GetSource(m_nSuctionSample).Stop();

			SetState(TransitionStage_MoveStage);
		}
	}
	else if (m_currentPlayState == TransitionStage_MoveStage)
	{
		int nDirection = (int)((m_stageWinner == Player::Blue) ? Stage::Stage_BlueDirection : Stage::Stage_RedDirection);
		Stage nextStage = (Stage)((int)m_currentStage + nDirection);
		m_currentStage = nextStage;
		if (nextStage <= Stage::Stage_BlueWin || nextStage >= Stage::Stage_RedWin)
		{
			SetState(Winning);
		}
		else
		{
			SetState(TransitionStage_BuildLevel);
		}
	}
	else if (m_currentPlayState == TransitionStage_BuildLevel)
	{
		// TODO: Fade out (as a pre state)

		// Strip out the current level
		// Push in a new one	
		RebuildLevel();

		// TODO: Fade in (as a post state)

		// Exhale
		SetState(TransitionStage_Exhale);
	}
	else if (m_currentPlayState == TransitionStage_Exhale)
	{
		if (GGJ2015::Get()->GetEntityManager()->IsSystemAtRest())
		{
			m_stageWinner = Player::None;
			EndTurn();
		}
	}
	else if (m_currentPlayState == Winning)
	{
		// Just delay here for a bit to let the dust settle then transition back to the menu state

		// TODO: Disable all hazards!

		// Only the winners can go.
		SuckSpacemenTowardStageDoor(m_stageWinner);

		std::vector<Spaceman*> spacemen;
		GGJ2015::Get()->GetEntityManager()->GetEntites<Spaceman>(spacemen);
		bool bAllWinnersSafe = true;
		for (auto spaceman : spacemen)
		{
			if (spaceman->m_player == m_stageWinner)
			{
				bAllWinnersSafe = false;
				break;
			}
		}

		// TODO:
		if (bAllWinnersSafe)
		{
			// PARTY MUSIC
		}

		if (m_fStageTimer > ms_fWinningTimeMs)
		{
			// return to the main menu
			GGJ2015::Get()->GetGameStateManager()->PushGameState<MainMenuState>();
			return;
		}
	}
}

void PlayState::Render()
{
	GGJ2015::Get()->GetCamera().Look();
	gRenderer->SetProjectionMode(Mode3D);
	gRenderer->UpdateModelviewMatrix();
	gRenderer->UpdateProjectionMatrix();

	gRenderer->SetBlendMode(ModeNormal);
	gRenderer->SetColor(cmColor(1.0f, 1.0f, 1.0f, 1.0f));
	gRenderer->EnableDepthMask(true);
	gRenderer->EnableDepthTest(true);
	gRenderer->BindTexture(gRenderer->GetTexture(GGJ2015::Get()->GetSceneTexture()).nId);

	
	cmFlatShader::Get().Activate();
	GGJ2015::Get()->GetSceneMatteMesh().Render();
	//GGJ2015::Get()->GetSceneDetailMesh().Render();

	//gRenderer->SetColor(cmColor(1.0f, 0.0f, 1.0f, 0.5f));
	//gRenderer->BindTexture(gRenderer->GetTexture(L"TEX_WHITE").nId);
	//GGJ2015::Get()->GetSceneCollisionMesh().Render();
	
		
	gRenderer->BindTexture(gRenderer->GetTexture(GGJ2015::Get()->GetSceneTexture()).nId);
	gRenderer->SetColor(cmColor(1.0f, 1.0f, 1.0f, 1.0f));
	gRenderer->UpdateModelviewMatrix();
	gRenderer->UpdateProjectionMatrix();

	cmPhongShader::Get().Activate();
	cmPhongShader::Get().SetLightPos(GGJ2015::Get()->GetLightPos() * GGJ2015::Get()->GetCamera().GetTransform().viewMatrix());
	cmPhongShader::Get().SetAmbient(0.4f);
	cmPhongShader::Get().SetSpecularColor(cmColor());
	cmPhongShader::Get().SetSpecularGloss(45.0f);
	cmPhongShader::Get().SetSpecularAmount(0.2f);
	cmPhongShader::Get().SetModelViewMatrix(gRenderer->GetModelViewMatrix());
	cmPhongShader::Get().SetNormalMatrix(gRenderer->GetNormalMatrix());
	GGJ2015::Get()->GetSceneSpecularMesh().Render();

// 	cmAlphaSpecShader::Get().Activate();
// 	cmAlphaSpecShader::Get().SetLightPos(GGJ2015::Get()->GetLightPos() * GGJ2015::Get()->GetCamera().GetTransform().viewMatrix());
// 	cmAlphaSpecShader::Get().SetSpecularColor(cmColor());
// 	cmAlphaSpecShader::Get().SetSpecularGloss(45.0f);
// 	cmAlphaSpecShader::Get().SetSpecularAmount(0.2f);
// 	cmAlphaSpecShader::Get().SetModelViewMatrix(gRenderer->GetModelViewMatrix());
// 	cmAlphaSpecShader::Get().SetNormalMatrix(gRenderer->GetNormalMatrix());
// 	GGJ2015::Get()->GetSceneAlphaMesh().Render();

	GGJ2015::Get()->GetEntityManager()->Render();

	//cmLocatorManager::Get().Render();

	cmScreenManager::Get().Render();
}

bool PlayState::OnMouseDown(const cmMouseButtons& button, const Vector3& vPos)
{
	bDragging = true;
	// is the position close to a spaceman
	// select the spaceman for launch

	if (m_currentPlayState == Select || m_currentPlayState == PrepareLaunch)
	{
		// You can select a new spaceman instead of launching

		//get vPos in worlds space
		Vector3 vIntersection;
		if (GetScreenSpaceMouseWorldIntersect(vIntersection))
		{
			std::vector<Spaceman*> spacemen;
			GGJ2015::Get()->m_pEntityManager->GetEntites<Spaceman>(spacemen);
			for (auto spaceman : spacemen)
			{
				vIntersection.z = spaceman->m_transform.position.z;
				float fDistance = spaceman->m_transform.position.distance(vIntersection);
				if (fDistance < spaceman->ms_sphereRadius && spaceman->m_player == GGJ2015::Get()->m_playerTurn)
				{
					SetSpaceman(spaceman);
					break;
				}
			}
		}

		m_touchPosition = vPos;
		return true;
	}
	else if (m_currentPlayState == Launched && m_nNumPunches < ms_nMaxPunches && m_pSelectedSpaceman->m_fPunchTimer <= 0.0f)
	{
		// SPAAACE PUUUNCH! (On mousedown for immediacy)

		Vector3 vPunchDirection;
		float fMinDistance = 999999.0f;
		Vector3 vCurrentSpacemanPosition = m_pSelectedSpaceman->m_transform.position;
		std::vector<Spaceman*> spacemen;
		GGJ2015::Get()->m_pEntityManager->GetEntites<Spaceman>(spacemen);
		for (auto spaceman : spacemen)
		{
			if (spaceman == m_pSelectedSpaceman) // Can't spacepunch self..
				continue;

			vPunchDirection = spaceman->m_transform.position - vCurrentSpacemanPosition;
			float fDistance = vPunchDirection.magnitude();
			vPunchDirection = vPunchDirection.normalize();

			if (fDistance - spaceman->ms_sphereRadius < m_pSelectedSpaceman->ms_punchRadius)// && fDistance < fMinDistance)
			{
				fMinDistance = fDistance;

				//PUNCH 'spaceman'
				++m_nNumPunches;

				vPunchDirection.z = 0.0f;
				spaceman->Launch(vPunchDirection, m_pSelectedSpaceman->ms_punchPower);
				int punchId = 1 + (m_pSelectedSpaceman->m_rng.Get() % 3);
				std::wstring punchSample(L"Hit " + toStringW(punchId) + L".ogg");
				gAudio->Play(punchSample, 0.7f);

				double angleBetween = AngleBetweenVectorsXY(m_pSelectedSpaceman->m_transform.forward, vPunchDirection);
				angleBetween /= PI;

				if (angleBetween > 1.0f)
					angleBetween = 2.0f - angleBetween;
				if (angleBetween < -1.0f)
					angleBetween = 2.0f + angleBetween;

				Spaceman::ePunchDirection punchDir = Spaceman::ePunchDirection::None;
				if (angleBetween > 0.5f)
					punchDir = Spaceman::ePunchDirection::BottomRight;
				else if (angleBetween > 0.0f)
					punchDir = Spaceman::ePunchDirection::TopRight;
				else if (angleBetween < -0.5f)
					punchDir = Spaceman::ePunchDirection::BottomLeft;
				else
					punchDir = Spaceman::ePunchDirection::TopLeft;
				
				m_pSelectedSpaceman->Punch(punchDir);

				//printf("PUNCH: Q%d - angle:%.2f\n", m_pSelectedSpaceman->m_ePunchDirection, angleBetween);
				break;
			}
		}
	}

	return false;
}

bool PlayState::OnMouseMove(const Vector3& vPos)
{
	if (m_currentPlayState == PrepareLaunch && bDragging)
	{
		// update and draw our ui
		return true;
	}

	return false;
}

bool PlayState::OnMouseUp(const cmMouseButtons& button, const Vector3& vPos)
{
	bDragging = false;
	
	if (m_currentPlayState < Launched)
	{
		if (m_pSelectedSpaceman)
		{
			Vector3 vIntersection;
			GetScreenSpaceMouseWorldIntersect(vIntersection);

			// if we released a launch distance away, then fire our spaceman
			vIntersection.z = m_pSelectedSpaceman->m_transform.position.z;
			float fSpacemanDistance = m_pSelectedSpaceman->m_transform.position.distance(vIntersection);

			// check to see if we've dragged far enough away
			float fLaunchPower = CalculateLaunchPower(vPos);
			if (fLaunchPower  > 0.0f)
			{
				// FIRE ZE SPACEMAN
				bool bInverse = true;

				Vector3 direction = bInverse ? (m_touchPosition - vPos) : (vPos - m_touchPosition);

				// Rotate the vector 90 degrees
				direction = RotatePointAroundVector(direction, Vector3::up(), FHALF_PI);
				// Invert y component
				direction.y *= -1.0f;

				gAudio->Play(L"Launch - Fire.ogg");
				m_pSelectedSpaceman->Launch(direction.normalize(), fLaunchPower);
				
				// clamp to max
				SetState(Launched);
				m_nNumPunches = 0;
			}
			else if (fSpacemanDistance > m_pSelectedSpaceman->ms_sphereRadius)
			{
				// if the touch positon was outside the the spaceman touch threshold
				// and outside the firing distance. then just release the spaceman
				ClearSelectedSpaceman();
			}
		}
	}
// 	else if (m_currentPlayState == Launched)
// 	{
// 		// TODO: If spaceman is flying we should be punching if its available
// 		m_nNumPunches++;
// 	}
	
	return true;
}

float PlayState::CalculateLaunchPower(const Vector3& releasePosition)
{
	float fScreenDragDistance = m_touchPosition.distance(releasePosition);

	fScreenDragDistance = min(fScreenDragDistance, ms_fMaximumScreenDragDistance);

	// remove min. if it goes negative thats fine
	fScreenDragDistance -= ms_fMinimumScreenDragDistance;

	// get a percentage of the power we're allowing
	fScreenDragDistance /= (ms_fMaximumScreenDragDistance - ms_fMinimumScreenDragDistance);

	return fScreenDragDistance;
}

bool PlayState::GetScreenSpaceMouseWorldIntersect(Vector3& vIntersection)
{
	GGJ2015::Get()->GetCamera().Look();
	gRenderer->SetProjectionMode(Mode3D);
	gRenderer->UpdateModelviewMatrix();
	gRenderer->UpdateProjectionMatrix();
	gRenderer->UpdateMouseVectors();
	Vector3 tempface[4];
	tempface[0] = Vector3(0.0f, 0.0f, GGJ2015::Get()->GetZConstraint());
	tempface[1] = Vector3(1.0f, 0.0f, GGJ2015::Get()->GetZConstraint());
	tempface[2] = Vector3(1.0f, 1.0f, GGJ2015::Get()->GetZConstraint());
	tempface[3] = Vector3(0.0f, 1.0f, GGJ2015::Get()->GetZConstraint());

	return RayPickPlane(tempface, vIntersection);
}

bool PlayState::OnTouchStart()
{
	return true;
}

bool PlayState::OnTouchEnd()
{
	return true;
}

void PlayState::SetSpaceman(Spaceman* pSpaceman)
{
	m_pSelectedSpaceman = pSpaceman;
	SetState(PrepareLaunch);
	// TODO: Select in the ui
}

void PlayState::ClearSelectedSpaceman()
{
	SetState(Select);
	m_nNumPunches = 0;
	m_pSelectedSpaceman = nullptr;
}

void PlayState::EndTurn()
{
	ChangePlayer();
	ClearSelectedSpaceman(); // should set the stage back to select.
	SetState(Respawn);	// So we'll transition to respawn if the player needs it
}

void PlayState::ChangePlayer()
{
	GGJ2015::Get()->SetPlayerTurn((GGJ2015::Get()->m_playerTurn == Player::Blue) ? Player::Red : Player::Blue);
}

void PlayState::RespawnSpaceman(Player owner)
{
	SpawnChute* pChute = (owner == Player::Blue) ? m_pBlueSpawnChute : m_pRedSpawnChute;
	pChute->SpawnSpaceman(owner);
	gAudio->Play(L"Ejection.ogg");
}

void PlayState::RebuildLevel(bool bInitial)
{
	ClearSelectedSpaceman();

	// Remove all entities
	std::vector<std::shared_ptr<Entity> > carryover;
	if (!bInitial)
	{
		carryover = (m_stageWinner == Player::Blue) ? m_pRedDoor->m_transitioningSpacemen : m_pBlueDoor->m_transitioningSpacemen;
	}

	GGJ2015::Get()->GetEntityManager()->Purge();

	// Load the blast doors
	// becuase artists
	m_pRedDoor = GGJ2015::Get()->GetEntityManager()->SpawnEntity<BlastDoor>("door_locator_blue", Player::Red);
	m_pBlueDoor = GGJ2015::Get()->GetEntityManager()->SpawnEntity<BlastDoor>("door_locator_red", Player::Blue);
	
	// Load the spawners
	m_pRedSpawnChute = GGJ2015::Get()->GetEntityManager()->SpawnEntity<SpawnChute>("chute_red001");
	m_pBlueSpawnChute = GGJ2015::Get()->GetEntityManager()->SpawnEntity<SpawnChute>("chute_blue001");
	
	// Load the hazards
	int nCurrentHazard = 0;
	char location[128];
	char trigger[128];
	while (++nCurrentHazard <= 20)
	{
		sprintf_s(location, "hazard_locator_%03d", nCurrentHazard);
		sprintf_s(trigger, "hazard_trigger_%03d", nCurrentHazard);
		if (cmLocatorManager::Get().DoesLocatorExist(location)
			&& cmLocatorManager::Get().DoesTriggerExist(trigger))
		{
			// Spawn a hazard
			SpawnHazard(cmLocatorManager::Get().GetLocator(location), trigger);
		}
		else
		{
			break;
		}
	}

	// TODO: Load the scenery (ask matt)
	
	// If this is initial then load the default spacemen
	float z = GGJ2015::Get()->GetZConstraint();
//	if (bInitial)
	{
		// Spawn initial spacemen
		for (int i = 0; i < 3; ++i)
		{
			float fAngle = ((1.0f + i) / 4.0f) * FPI;

			float x = sin(fAngle) * 10.0f;
			float y = cos(fAngle) * 20.0f;

			Spaceman* pred = GGJ2015::Get()->GetEntityManager()->SpawnEntity<Spaceman>(Player::Red);
			pred->m_transform.position = Vector3(x, y, GGJ2015::Get()->GetZConstraint());

			Spaceman* pblu = GGJ2015::Get()->GetEntityManager()->SpawnEntity<Spaceman>(Player::Blue);
			pblu->m_transform.position = Vector3(-x, y, GGJ2015::Get()->GetZConstraint());
		}
	}
	//else
	{
		// prepare the ejection of the spacemen?
	}
}

void PlayState::SpawnHazard(cmTransform& transform, const std::string& sTrigger)
{
	enum class HazardType : int
	{
		None,
		Flame,
		NUM_HAZARD_TYPES,
	};

	HazardType hazard =  (HazardType)(m_hazardRNG.Get() % (int)HazardType::NUM_HAZARD_TYPES);

	// Add spawners for various hazard types
	switch (hazard)
	{
	case HazardType::Flame:
	{
		
		GGJ2015::Get()->GetEntityManager()->SpawnEntity<FlameHazard>(transform, sTrigger);
		break;
	}
	case HazardType::None:
	default:
		break;
	}
}

void PlayState::SetPlayerStageWinner(Player winner)
{
	// set the turn to the winner so we can track him
	// also means the other player gets first go in the next stage
	if (m_stageWinner == Player::None)
	{
		m_stageWinner = winner;
		m_nSuctionSample = gAudio->Play(L"Airlock Suction.ogg");
		SetState(TransitionStage_Vaccuum);
	}
}

void PlayState::SuckSpacemenTowardStageDoor(Player player)
{
	// start sucking the players towards the correct door.
	BlastDoor* pDoor = m_stageWinner == Player::Blue ? m_pBlueDoor : m_pRedDoor;
	Vector3 doorPosition;
	if (pDoor)
	{
		doorPosition = pDoor->m_transform.position;
	}

	SuckSpacemenToward(doorPosition, player);
}


void PlayState::SuckSpacemenToward(const Vector3& position, Player player)
{
	float fElapsedSeconds = gPlatform->GetDeltaTime() / 1000.0f;

	std::vector<Spaceman*> spacemen;
	GGJ2015::Get()->GetEntityManager()->GetEntites<Spaceman>(spacemen);
	for (auto spaceman : spacemen)
	{
		if (player != Player::None && spaceman->m_player != player)
			continue; // not this guy

		// TODO: Calculate a direction and an increment to their velocities
		Vector3 direction = (position - spaceman->m_transform.position).normalize();
		spaceman->m_velocity += (direction * (ms_fVacuumAcceleration * fElapsedSeconds));
		spaceman->m_velocity.x = clampFloat(spaceman->m_velocity.x, -ms_fMaximumVacuumVelocity, ms_fMaximumVacuumVelocity);
		spaceman->m_velocity.y = clampFloat(spaceman->m_velocity.y, -ms_fMaximumVacuumVelocity, ms_fMaximumVacuumVelocity);
		spaceman->m_velocity.z = clampFloat(spaceman->m_velocity.z, -ms_fMaximumVacuumVelocity, ms_fMaximumVacuumVelocity);
	}
}